#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Logging"""

import logging
import logging.handlers

import utils


def get_logger(name="Pyhole"):
    """Log handler"""
    debug_option = utils.get_option("debug")
    debug_config = utils.get_config().get("debug", type="bool")
    debug = debug_option or debug_config

    log_dir = utils.get_directory("logs")
    log_level = logging.DEBUG if debug else logging.INFO
    log_format = "%(asctime)s [%(name)s] %(message)s"
    log_datefmt = "%H:%M:%S"

    logging.basicConfig(level=log_level, format=log_format,
            datefmt=log_datefmt)

    log = logging.handlers.TimedRotatingFileHandler("%s/%s.log" % (log_dir,
            name.lower()), "midnight")
    log.setLevel(log_level)
    formatter = logging.Formatter(log_format, log_datefmt)
    log.setFormatter(formatter)
    logging.getLogger(name).addHandler(log)

    return logging.getLogger(name)



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Entertainment Plugin"""

import re

from BeautifulSoup import BeautifulSoup

from pyhole import plugin
from pyhole import utils


class Entertainment(plugin.Plugin):
    """Provide access to entertaining sites"""

    @plugin.hook_add_command("grouphug")
    @utils.spawn
    def grouphug(self, params=None, **kwargs):
        """Display a random Group Hug (ex: .grouphug)"""
        url = "http://grouphug.us/random"
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        soup = BeautifulSoup(response.read())
        grouphug = utils.decode_entities(
                soup.findAll(id=re.compile("node-\d+"))[2].p.contents[0])
        self.irc.reply(grouphug)

    @plugin.hook_add_command("lastnight")
    @utils.spawn
    def lastnight(self, params=None, **kwargs):
        """Display a random Text From Last Night (ex: .lastnight)"""
        url = ("http://www.textsfromlastnight.com/"
                "Random-Texts-From-Last-Night.html")
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        soup = BeautifulSoup(response.read())
        lastnight = utils.decode_entities(
                soup.findAll(href=re.compile(
                        "/Text-Replies-\d+.html"))[0].contents[0])
        self.irc.reply(lastnight)



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Kernel.org Plugin"""

import re
import urllib

from BeautifulSoup import BeautifulSoup

from pyhole import plugin
from pyhole import utils


class Kernel(plugin.Plugin):
    """Provide access to kernel.org data"""

    @plugin.hook_add_command("kernel")
    @utils.spawn
    def kernel(self, params=None, **kwargs):
        """Retrieve the current kernel version (ex: .kernel)"""
        url = "http://kernel.org/kdist/finger_banner"
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        r = re.compile("(.* mainline .*)")
        m = r.search(response.read())
        kernel = m.group(1).replace("  ", "")
        self.irc.reply(kernel)

    @plugin.hook_add_keyword("k")
    @utils.spawn
    def keyword_k(self, params=None, **kwargs):
        """Retrieve kernel.org Bugzilla bug information (ex: K12345)"""
        if params:
            params = utils.ensure_int(params)
            if not params:
                return

            query = urllib.urlencode({"id": params})
            url = "http://bugzilla.kernel.org/show_bug.cgi?%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response or not isinstance(params, int):
                return

            soup = BeautifulSoup(response.read())
            desc = utils.decode_entities(soup.head.title.string)

            try:
                status = soup.find("span", {"id":
                        "static_bug_status"}).string.strip().capitalize()
                assignee = utils.decode_entities(soup.findAll("span",
                        {"class": "vcard"})[0].contents[0].string)

                self.irc.reply("Kernel.org %s [Status: %s, Assignee: %s] %s" %
                        (desc, status, assignee, url))
            except TypeError:
                return

    @plugin.hook_add_msg_regex(
            "https?:\/\/bugzilla\.kernel\.org\/show\_bug\.cgi\?id\=")
    def _watch_for_k_bug_url(self, params=None, **kwargs):
        """Watch for kernel.org Bugzilla bug URLs"""
        try:
            bug_id = kwargs["full_message"].split("id=", 1)[1].split(" ", 1)[0]
            self.keyword_k(bug_id)
        except TypeError:
            return



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Utility Library"""

from __future__ import with_statement

import eventlet
import os
import optparse
import re

from BeautifulSoup import BeautifulStoneSoup

import config
import version


eventlet.monkey_patch()


def admin(func):
    """Administration Decorator"""
    def wrap(self, *args, **kwargs):
        if self.irc.source in self.irc.admins:
            func(self, *args, **kwargs)
        else:
            self.irc.reply("Sorry, you are not authorized to do that.")
    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    wrap.__module__ = func.__module__
    return wrap


def spawn(func):
    """Greenthread Spawning Decorator"""
    def wrap(self, *args, **kwargs):
        eventlet.spawn_n(func, self, *args, **kwargs)
    wrap.__doc__ = func.__doc__
    wrap.__name__ = func.__name__
    wrap.__module__ = func.__module__
    return wrap


def decode_entities(html):
    """Strip HTML entities from a string and make it printable"""
    html = "".join(str(x).strip() for x in BeautifulStoneSoup(html,
            convertEntities=BeautifulStoneSoup.HTML_ENTITIES).findAll(
            text=True))

    return filter(lambda x: ord(x) > 9 and ord(x) < 127, html)


def ensure_int(param):
    """Ensure the given param is an int"""
    try:
        param = re.sub("\W", "", param)
        return int(param)
    except ValueError:
        return None


def build_options():
    """Generate command line options"""
    parser = optparse.OptionParser(version=version.version_string())
    parser.add_option("-c", "--config", default=get_conf_file_path(),
            help="specify the path to a configuration file")
    parser.add_option("-d", "--debug", action="store_true",
            help="show debugging output")

    return parser.parse_args()


def get_option(option):
    """Retrive an option from the command line."""
    options, _args = build_options()
    return vars(options).get(option)


def get_home_directory():
    """Return the home directory"""
    home_dir = os.getenv("HOME") + "/.pyhole/"
    if not os.path.exists(home_dir):
        os.makedirs(home_dir)

    return home_dir


def get_directory(new_dir):
    """Return a directory"""
    home_dir = get_home_directory()
    new_dir = home_dir + new_dir

    if not os.path.exists(new_dir):
        os.makedirs(new_dir)

    return new_dir + "/"


def get_conf_file_path():
    """Return the path to the conf file"""
    return get_home_directory() + "pyhole.conf"


def get_conf_file():
    """Return the path to the conf file"""
    return get_option("config") or get_conf_file_path()


def get_config(section="Pyhole"):
    """Return the default config object"""
    return config.Config(get_conf_file(), section)


def write_file(directory, file_name, data):
    """Write data to file"""
    directory = get_directory(directory)
    with open(directory + file_name, "w") as open_file:
        open_file.write(str(data).strip())


def read_file(directory, file_name):
    """Read and return the data in file"""
    directory = get_directory(directory)
    try:
        with open(directory + file_name, "r") as open_file:
            data = open_file.read()
        return data
    except IOError:
        return None


def generate_config():
    """Generate an example config"""
    example = """# Global Configuration

[Pyhole]
admins: nick!ident, nick2!ident
command_prefix: .
reconnect_delay: 60
rejoin_delay: 5
debug: False
plugins: admin, dice, entertainment, news, search, urls, weather

[Redmine]
domain: redmine.example.com
key: abcd1234

# IRC Network Configuration

[FreeNode]
server: verne.freenode.net
password:
port: 7000
ssl: True
ipv6: True
bind_to: fe80::1
nick: mynick
identify_password: mypass
channels: #mychannel key, #mychannel2

[EFnet]
server: irc.efnet.net
password:
port: 6667
ssl: False
ipv6: False
bind_to:
nick: mynick
identify_password:
channels: #mychannel key, #mychannel2
"""

    conf_file = get_conf_file()
    if os.path.exists(conf_file):
        return

    print "Generating..."
    with open(conf_file, "w") as open_file:
        open_file.write(example)
    print "Done"



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole URL Plugin"""

from BeautifulSoup import BeautifulSoup

from pyhole import plugin
from pyhole import utils


class Url(plugin.Plugin):
    """Provide access to URL data"""

    def __init__(self, irc):
        self.irc = irc
        self.name = self.__class__.__name__
        self.url = None

    @plugin.hook_add_command("title")
    @utils.spawn
    def title(self, params=None, **kwargs):
        """Display the title of the a URL (ex: .title <url>)"""
        if params:
            self._find_title(params.split(" ", 1)[0])
        else:
            if self.url:
                self._find_title(self.url)

    @plugin.hook_add_msg_regex("https?:\/\/|www\.")
    def _watch_for_url(self, params=None, **kwargs):
        """Watch and keep track of the latest URL"""
        try:
            self.url = kwargs["full_message"].split(" ", 1)[0]
        except TypeError:
            return

    def _find_title(self, url):
        """Find the title of a given URL"""
        if not url.startswith("http://"):
            url = "http://" + url

        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        soup = BeautifulSoup(response.read())

        if soup.head:
            title = utils.decode_entities(soup.head.title.string)
            self.irc.reply(title)
        else:
            self.irc.reply("No title found for %s" % url)



#   Copyright 2011 Paul Voccio
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Cleverbot Plugin"""

import cleverbot

from pyhole import plugin
from pyhole import utils


class CleverChat(plugin.Plugin):
    """Provide access to the Cleverbot API"""

    @plugin.hook_add_command("cleverbot")
    @utils.spawn
    def cleverbot(self, params=None, **kwargs):
        """Chat with the Cleverbot (ex: .chat 'How you doin'?"""
        cb = cleverbot.Session()
        if params:
            try:
                self.irc.reply(cb.Ask(params))
            except KeyError:
                self.irc.reply("Can't chat now. I'm washing my hair.")
        else:
            self.irc.reply(self.chat.__doc__)

    @plugin.hook_add_command("cb")
    def alias_cb(self, params=None, **kwargs):
        """Alias of cleverbot"""
        self.cleverbot(params, **kwargs)



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole News Plugin"""

from xml.dom import minidom

from pyhole import plugin
from pyhole import utils


class News(plugin.Plugin):
    """Provide access to news feeds"""

    @plugin.hook_add_command("cnn")
    @utils.spawn
    def cnn(self, params=None, **kwargs):
        """Display current CNN news (ex: .cnn)"""
        url = "http://rss.cnn.com/rss/cnn_topstories.rss"
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        xml = minidom.parseString(response.read())
        for i, item in enumerate(xml.childNodes[2].childNodes[0].childNodes):
            if i >= 21 and i <= 28 and item.childNodes:
                ref = item.childNodes
                self.irc.reply("%s: %s" % (ref[1].firstChild.data,
                        ref[5].firstChild.data))

    @plugin.hook_add_command("digg")
    @utils.spawn
    def digg(self, params=None, **kwargs):
        """Display current Digg news (ex: .digg)"""
        url = "http://services.digg.com/2.0/story.getTopNews?type=rss"
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        xml = minidom.parseString(response.read())
        for i, item in enumerate(xml.childNodes[0].childNodes[1].childNodes):
            if i >= 15 and i <= 21 and item.childNodes:
                ref = item.childNodes
                self.irc.reply("%s: %s" % (ref[1].firstChild.data.strip(),
                        ref[3].firstChild.data.strip()))

    @plugin.hook_add_command("reddit")
    @utils.spawn
    def reddit(self, params=None, **kwargs):
        """Display current reddit news (ex: .reddit)"""
        url = "http://www.reddit.com/.rss"
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        xml = minidom.parseString(response.read())
        for i, item in enumerate(xml.childNodes[0].childNodes[0].childNodes):
            if i >= 4 and i <= 7:
                ref = item.childNodes
                self.irc.reply("%s: %s" % (
                        ref[0].firstChild.data.encode("ascii", "ignore"),
                        ref[1].firstChild.data.encode("ascii", "ignore")))



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Weather Plugin"""

import pywapi

from pyhole import plugin
from pyhole import utils


class Weather(plugin.Plugin):
    """Provide access to current weather data"""

    @plugin.hook_add_command("weather")
    @utils.spawn
    def weather(self, params=None, **kwargs):
        """Display current weather report (ex: .w [set] [<location>])"""
        if params:
            location = params
            if location.startswith("set "):
                location = location[4:]
                utils.write_file(self.name, self.irc.source, location)
                self.irc.reply("Location information saved")
        else:
            location = utils.read_file(self.name, self.irc.source)
            if not location:
                self.irc.reply(self.weather.__doc__)
                return

        try:
            w = pywapi.get_weather_from_google(location)
        except Exception:
            self.irc.reply("Unable to fetch weather data")
            return

        if w["current_conditions"]:
            city = w["forecast_information"]["city"]
            temp_f = w["current_conditions"]["temp_f"]
            temp_c = w["current_conditions"]["temp_c"]
            humidity = w["current_conditions"]["humidity"]
            wind = w["current_conditions"]["wind_condition"]
            condition = w["current_conditions"]["condition"]

            result = "%s: %sF/%sC   %s   %s   %s" % (city, temp_f, temp_c,
                    humidity, wind, condition)
            self.irc.reply(result)
        else:
            self.irc.reply("Location not found: '%s'" % location)

    @plugin.hook_add_command("w")
    def alias_w(self, params=None, **kwargs):
        """Alias of weather"""
        self.weather(params, **kwargs)



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Version Handling"""

from __future__ import with_statement

import os
import sys


__VERSION__ = "0.6.0"


def current_git_hash():
    """Return the current git hash"""
    git_file = ".git/refs/heads/master"
    git_path = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
            os.pardir, os.pardir, git_file))

    if not os.path.exists(git_path):
        git_path = os.getcwd() + "/" + git_file
        if not os.path.exists(git_path):
            git_path = os.getcwd() + "/../" + git_file
            if not os.path.exists(git_path):
                return None

    with open(git_path, "r") as git:
        git_hash = git.read()

    return git_hash[0:5]


def version_string():
    """Return the full version"""
    git_hash = current_git_hash()
    if git_hash:
        return "pyhole v%s (%s) - http://pyhole.org" % (__VERSION__, git_hash)

    return "pyhole v%s - http://pyhole.org" % __VERSION__


def version_hash():
    """Return the current version with git hash"""
    git_hash = current_git_hash()
    return "%s-%s" % (__VERSION__, git_hash)


def version():
    """Return the current version"""
    return __VERSION__



#   Copyright 2011 John Dickinson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Dice Plugin"""

import random

from pyhole import plugin


class Dice(plugin.Plugin):
    """Provide access to dice games"""

    @plugin.hook_add_command("roll")
    def roll(self, params=None, **kwargs):
        """Roll dice (ex: .roll 2d6)"""
        if params:
            query = params.split()[0]
            try:
                if "d" not in query:
                    count = 1
                    sides = int(query)
                elif query.startswith("d"):
                    count = 1
                    sides = int(query[1:])
                else:
                    count, sides = query.split("d")
                    count = int(count)
                    sides = int(sides)
            except (TypeError, ValueError):
                result = self.roll.__doc__
            else:
                result = 0
                for _ in xrange(count):
                    result += random.randint(1, sides)
        else:
            result = self.roll.__doc__

        self.irc.reply(result)



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Event-based IRC Class"""

import multiprocessing
import random
import re
import sys
import time
import urllib

import irclib
import log
import plugin
import utils
import version


LOG = log.get_logger()
CONFIG = utils.get_config()


class IRC(irclib.SimpleIRCClient):
    """An IRClib connection."""

    def __init__(self, network):
        irclib.SimpleIRCClient.__init__(self)
        network_config = utils.get_config(network)

        self.log = log.get_logger(str(network))
        self.version = version.version_string()
        self.source = None
        self.target = None
        self.addressed = False

        self.admins = CONFIG.get("admins", type="list")
        self.command_prefix = CONFIG.get("command_prefix")
        self.reconnect_delay = CONFIG.get("reconnect_delay", type="int")
        self.rejoin_delay = CONFIG.get("rejoin_delay", type="int")

        self.server = network_config.get("server")
        self.password = network_config.get("password", default="")
        self.port = network_config.get("port", type="int", default=6667)
        self.ssl = network_config.get("ssl", type="bool", default=False)
        self.ipv6 = network_config.get("ipv6", type="bool", default=False)
        self.bind_to = network_config.get("bind_to", default="")
        self.nick = network_config.get("nick")
        self.identify_password = network_config.get("identify_password",
                default="")
        self.channels = network_config.get("channels", type="list")

        self.load_plugins()

        self.log.info("Connecting to %s:%d as %s" % (self.server, self.port,
                self.nick))
        self.connect(self.server, self.port, self.nick, self.password,
                ssl=self.ssl, ipv6=self.ipv6, localaddress=self.bind_to)

    def load_plugins(self, reload_plugins=False):
        """Load plugins and their commands respectively."""
        if reload_plugins:
            plugin.reload_plugins(irc=self)
        else:
            plugin.load_plugins(irc=self)

        self.log.info("Loaded Plugins: %s" % active_plugins())

    def run_hook_command(self, mod_name, func, arg, **kwargs):
        """Make a call to a plugin hook."""
        try:
            func(arg, **kwargs)
            if arg:
                self.log.debug("Calling: %s.%s(\"%s\")" % (mod_name,
                        func.__name__, arg))
            else:
                self.log.debug("Calling: %s.%s(None)" % (mod_name,
                        func.__name__))
        except Exception, exc:
            self.log.error(exc)

    def run_msg_regexp_hooks(self, message, private):
        """Run regexp hooks."""
        for mod_name, func, msg_regex in plugin.hook_get_msg_regexs():
            match = re.search(msg_regex, message, re.I)
            if match:
                self.run_hook_command(mod_name, func, match, private=private,
                        full_message=message)

    def run_keyword_hooks(self, message, private):
        """Run keyword hooks."""
        words = message.split(" ")
        for mod_name, func, kwarg in plugin.hook_get_keywords():
            for word in words:
                match = re.search("^%s(.+)" % kwarg, word, re.I)
                if match:
                    self.run_hook_command(mod_name, func, match.group(1),
                            private=private, full_message=message)

    def run_command_hooks(self, message, private):
        """Run command hooks."""
        for mod_name, func, cmd in plugin.hook_get_commands():
            self.addressed = False

            if private:
                match = re.search("^%s$|^%s\s(.*)$" % (cmd, cmd), message,
                        re.I)
                if match:
                    self.run_hook_command(mod_name, func, match.group(1),
                            private=private, addressed=self.addressed,
                            full_message=message)

            if message.startswith(self.command_prefix):
                # Strip off command prefix
                msg_rest = message[len(self.command_prefix):]
            else:
                # Check for command starting with nick being addressed
                msg_start_upper = message[:len(self.nick) + 1].upper()
                if msg_start_upper == self.nick.upper() + ":":
                    # Get rest of string after "nick:" and white spaces
                    msg_rest = re.sub("^\s+", "",
                            message[len(self.nick) + 1:])
                else:
                    continue

                self.addressed = True

            match = re.search("^%s$|^%s\s(.*)$" % (cmd, cmd), msg_rest, re.I)
            if match:
                self.run_hook_command(mod_name, func, match.group(1),
                        private=private, addressed=self.addressed,
                        full_message=message)

    def poll_messages(self, message, private=False):
        """Watch for known commands."""
        self.addressed = False

        self.run_command_hooks(message, private)
        self.run_keyword_hooks(message, private)
        self.run_msg_regexp_hooks(message, private)

    def reply(self, msg):
        """Send a privmsg."""
        if not hasattr(msg, "encode"):
            try:
                msg = str(msg)
            except Exception:
                self.log.error("msg cannot be converted to string")
                return

        msg = msg.encode("utf-8").split("\n")
        # 10 is completely arbitrary for now
        if len(msg) > 10:
            msg = msg[0:8]
            msg.append("...")

        for line in msg:
            if self.addressed:
                source = self.source.split("!")[0]
                self.connection.privmsg(self.target, "%s: %s" % (source, line))
                self.log.info("-%s- <%s> %s: %s" % (self.target, self.nick,
                        source, line))
            else:
                self.connection.privmsg(self.target, line)
                if irclib.is_channel(self.target):
                    self.log.info("-%s- <%s> %s" % (self.target, self.nick,
                            line))
                else:
                    self.log.info("<%s> %s" % (self.nick, line))

    def privmsg(self, target, msg):
        """Send a privmsg."""
        self.connection.privmsg(target, msg)

    def op_user(self, params):
        """Op a user."""
        params = params.split(" ", 1)
        self.connection.mode(params[0], "+o %s" % params[1])

    def deop_user(self, params):
        """De-op a user."""
        params = params.split(" ", 1)
        self.connection.mode(params[0], "-o %s" % params[1])

    def set_nick(self, params):
        """Set IRC nick."""
        self.nick = params
        self.connection.nick(params)

    def join_channel(self, params):
        """Join a channel."""
        channel = params.split(" ", 1)
        self.reply("Joining %s" % channel[0])
        if irclib.is_channel(channel[0]):
            self.channels.append(channel[0])
            if len(channel) > 1:
                self.connection.join(channel[0], channel[1])
            else:
                self.connection.join(channel[0])

    def part_channel(self, params):
        """Part a channel."""
        self.channels.remove(params)
        self.reply("Parting %s" % params)
        self.connection.part(params)

    def fetch_url(self, url, name):
        """Fetch a URL."""
        class PyholeURLopener(urllib.FancyURLopener):
            """Set a custom user agent."""
            version = self.version

        urllib._urlopener = PyholeURLopener()

        try:
            return urllib.urlopen(url)
        except IOError:
            self.reply("Unable to fetch %s data" % name)
            return None

    def on_nicknameinuse(self, connection, _event):
        """Ensure the use of unique IRC nick."""
        random_int = random.randint(1, 100)
        self.log.info("IRC nick '%s' is currently in use" % self.nick)
        self.nick = "%s%d" % (self.nick, random_int)
        self.log.info("Setting IRC nick to '%s'" % self.nick)
        connection.nick("%s" % self.nick)
        # Try to prevent nick flooding
        time.sleep(1)

    def on_welcome(self, connection, _event):
        """Join channels upon successful connection."""
        if self.identify_password:
            self.privmsg("NickServ", "IDENTIFY %s" % self.identify_password)

        for channel in self.channels:
            channel = channel.split(" ", 1)
            if irclib.is_channel(channel[0]):
                if len(channel) > 1:
                    connection.join(channel[0], channel[1])
                else:
                    connection.join(channel[0])

    def on_disconnect(self, _connection, _event):
        """Attempt to reconnect after disconnection."""
        self.log.info("Disconnected from %s:%d" % (self.server, self.port))
        self.log.info("Reconnecting in %d seconds" % self.reconnect_delay)
        time.sleep(self.reconnect_delay)
        self.log.info("Connecting to %s:%d as %s" % (self.server, self.port,
                self.nick))
        self.connect(self.server, self.port, self.nick, self.password,
                ssl=self.ssl)

    def on_kick(self, connection, event):
        """Automatically rejoin channel if kicked."""
        source = irclib.nm_to_n(event.source())
        target = event.target()
        nick, reason = event.arguments()

        if nick == self.nick:
            self.log.info("-%s- kicked by %s: %s" % (target, source, reason))
            self.log.info("-%s- rejoining in %d seconds" % (target,
                    self.rejoin_delay))
            time.sleep(self.rejoin_delay)
            connection.join(target)
        else:
            self.log.info("-%s- %s was kicked by %s: %s" % (target, nick,
                    source, reason))

    def on_invite(self, _connection, event):
        """Join a channel upon invitation."""
        source = event.source().split("@", 1)[0]
        if source in self.admins:
            self.join_channel(event.arguments()[0])

    def on_ctcp(self, connection, event):
        """Respond to CTCP events."""
        source = irclib.nm_to_n(event.source())
        ctcp = event.arguments()[0]

        if ctcp == "VERSION":
            self.log.info("Received CTCP VERSION from %s" % source)
            connection.ctcp_reply(source, "VERSION %s" % self.version)
        elif ctcp == "PING":
            if len(event.arguments()) > 1:
                self.log.info("Received CTCP PING from %s" % source)
                connection.ctcp_reply(source,
                        "PING %s" % event.arguments()[1])

    def on_join(self, _connection, event):
        """Handle joins."""
        target = event.target()
        source = irclib.nm_to_n(event.source())
        self.log.info("-%s- %s joined" % (target, source))

    def on_part(self, _connection, event):
        """Handle parts."""
        target = event.target()
        source = irclib.nm_to_n(event.source())
        self.log.info("-%s- %s left" % (target, source))

    def on_quit(self, _connection, event):
        """Handle quits."""
        source = irclib.nm_to_n(event.source())
        self.log.info("%s quit" % source)

    def on_action(self, _connection, event):
        """Handle IRC actions."""
        target = event.target()
        source = irclib.nm_to_n(event.source())
        msg = event.arguments()[0]
        self.log.info(unicode("-%s- * %s %s" % (target, source, msg), "utf-8"))

    def on_privnotice(self, _connection, event):
        """Handle private notices."""
        source = irclib.nm_to_n(event.source())
        msg = event.arguments()[0]
        self.log.info(unicode("-%s- %s" % (source, msg), "utf-8"))

    def on_pubnotice(self, _connection, event):
        """Handle public notices."""
        target = event.target()
        source = irclib.nm_to_n(event.source())
        msg = event.arguments()[0]
        self.log.info(unicode("-%s- <%s> %s" % (target, source, msg),
                "utf-8"))

    def on_privmsg(self, _connection, event):
        """Handle private messages."""
        self.source = event.source().split("@", 1)[0]
        self.target = irclib.nm_to_n(event.source())
        msg = event.arguments()[0]

        if self.target != self.nick:
            self.log.info(unicode("<%s> %s" % (self.target, msg), "utf-8"))
            self.poll_messages(msg, private=True)

    def on_pubmsg(self, _connection, event):
        """Handle public messages."""
        self.source = event.source().split("@", 1)[0]
        self.target = event.target()
        nick = irclib.nm_to_n(event.source())
        msg = event.arguments()[0]

        self.log.info(unicode("-%s- <%s> %s" % (self.target, nick, msg),
                "utf-8"))
        self.poll_messages(msg)


class IRCProcess(multiprocessing.Process):
    """An IRC network connection process."""
    def __init__(self, irc_network):
        super(IRCProcess, self).__init__()
        self.irc_network = irc_network
        self.reconnect_delay = CONFIG.get("reconnect_delay", type="int")

    def run(self):
        """An IRC network connection."""
        while True:
            try:
                connection = IRC(self.irc_network)
            except Exception, exc:
                LOG.error(exc)
                LOG.error("Retrying in %d seconds" % self.reconnect_delay)
                time.sleep(self.reconnect_delay)
                continue

            try:
                connection.start()
            except KeyboardInterrupt:
                sys.exit(0)
            except Exception, exc:
                LOG.error(exc)
                LOG.error("Retrying in %d seconds" % self.reconnect_delay)
                time.sleep(self.reconnect_delay)
                continue


def active_plugins():
    """List active plugins."""
    return ", ".join(sorted(plugin.active_plugins()))


def active_commands():
    """List active commands."""
    return ", ".join(sorted(plugin.active_commands()))


def active_keywords():
    """List active keywords."""
    return ", ".join(sorted(plugin.active_keywords()))


def network_list():
    """Return the list of IRC networks."""
    sections = CONFIG.sections()
    return [net for net in sections if net not in ["Pyhole", "Redmine"]]


def main():
    """Main IRC loop."""
    networks = network_list()

    LOG.info("Starting %s" % version.version_string())
    LOG.info("Connecting to IRC Networks: %s" % ", ".join(networks))
    for network in networks:
        proc = IRCProcess(network)
        proc.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        LOG.info("Caught KeyboardInterrupt, shutting down")



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Search Plugin"""

try:
    import json
except ImportError:
    import simplejson as json

import re
import urllib

from BeautifulSoup import BeautifulSoup
from xml.dom import minidom

from pyhole import plugin
from pyhole import utils


class Search(plugin.Plugin):
    """Provide access to search engines"""

    @plugin.hook_add_command("google")
    @utils.spawn
    def google(self, params=None, **kwargs):
        """Search Google (ex: .g <query>)"""
        if params:
            query = urllib.urlencode({"q": params})
            url = ("http://ajax.googleapis.com/ajax/"
                    "services/search/web?v=1.0&%s" % query)
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            json_obj = json.loads(response.read())
            results = json_obj["responseData"]["results"]
            if results:
                for r in results:
                    self.irc.reply("%s: %s" % (
                            r["titleNoFormatting"].encode("ascii", "ignore"),
                            r["unescapedUrl"]))
            else:
                self.irc.reply("No results found: '%s'" % params)
        else:
            self.irc.reply(self.google.__doc__)

    @plugin.hook_add_command("g")
    def alias_g(self, params=None, **kwargs):
        """Alias of google"""
        self.google(params, **kwargs)

    @plugin.hook_add_command("imdb")
    @utils.spawn
    def imdb(self, params=None, **kwargs):
        """Search IMDb (ex: .imdb <query>)"""
        if params:
            query = urllib.urlencode({"q": params})
            url = "http://www.imdb.com/find?s=all&%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            soup = BeautifulSoup(response.read())
            results = soup.findAll("td", {"valign": "top"})

            i = 0
            for result in results:
                if len(result) > 3 and len(result.contents[2].attrs) > 0:
                    id = result.contents[2].attrs[0][1]
                    title = utils.decode_entities(
                            result.contents[2].contents[0])
                    year = result.contents[2].nextSibling.strip()[0:6]

                    if not title.startswith("aka") and len(year):
                        self.irc.reply("%s %s: http://www.imdb.com%s" % (
                                title, year, id))
                        i += 1
                elif i >= 4:
                    break

            if i == 0:
                self.irc.reply("No results found: '%s'" % params)
        else:
            self.irc.reply(self.imdb.__doc__)

    @plugin.hook_add_command("twitter")
    @utils.spawn
    def twitter(self, params=None, **kwargs):
        """Search Twitter (ex: .twitter <query>)"""
        if params:
            query = urllib.urlencode({"q": params, "rpp": 4})
            url = "http://search.twitter.com/search.json?%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            json_obj = json.loads(response.read())
            results = json_obj["results"]
            if results:
                for r in results:
                    self.irc.reply("@%s: %s" % (r["from_user"],
                            utils.decode_entities(
                                    r["text"].encode("ascii", "ignore"))))
            else:
                self.irc.reply("No results found: '%s'" % params)
        else:
            self.irc.reply(self.twitter.__doc__)

    @plugin.hook_add_command("urban")
    @utils.spawn
    def urban(self, params=None, **kwargs):
        """Search Urban Dictionary (ex: .urban <query>)"""
        if params:
            query = urllib.urlencode({"term": params})
            url = "http://www.urbandictionary.com/define.php?%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            soup = BeautifulSoup(response.read())
            results = soup.findAll("div", {"class": "definition"})

            urban = ""
            if len(results):
                urban = " ".join(str(x) for x in soup.findAll(
                        "div", {"class": "definition"})[0].contents)

            if len(urban) > 0:
                for i, line in enumerate(urban.split("<br/>")):
                    if i <= 4:
                        self.irc.reply(utils.decode_entities(line))
                    else:
                        self.irc.reply("[...] %s" % url)
                        break
            else:
                self.irc.reply("No results found: '%s'" % params)
        else:
            self.irc.reply(self.urban.__doc__)

    @plugin.hook_add_command("wikipedia")
    @utils.spawn
    def wikipedia(self, params=None, **kwargs):
        """Search Wikipedia (ex: .wikipedia <query>)"""
        if params:
            query = urllib.urlencode({"action": "query",
                    "generator": "allpages", "gaplimit": 4,
                    "gapfrom": params, "format": "xml"})
            url = "http://en.wikipedia.org/w/api.php?%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            xml = minidom.parseString(response.read())
            for i in xml.childNodes[0].childNodes[1].childNodes[0].childNodes:
                title = i._attrs["title"].firstChild.data
                title = re.sub(" ", "_", title)
                self.irc.reply("http://en.wikipedia.org/wiki/%s" % title)
        else:
            self.irc.reply(self.wikipedia.__doc__)

    @plugin.hook_add_command("youtube")
    @utils.spawn
    def youtube(self, params=None, **kwargs):
        """Search YouTube (ex: .youtube <query>)"""
        if params:
            query = urllib.urlencode({"q": params, "v": 2, "max-results": 4,
                    "alt": "jsonc"})
            url = "http://gdata.youtube.com/feeds/api/videos?%s" % query
            response = self.irc.fetch_url(url, self.name)
            if not response:
                return

            json_obj = json.loads(response.read())
            results = json_obj["data"]
            if len(results) > 4:
                for r in results["items"]:
                    v = r["player"]["default"].split("&", 1)[0]
                    self.irc.reply("%s: %s" % (r["title"], v))
            else:
                self.irc.reply("No results found: '%s'" % params)
        else:
            self.irc.reply(self.youtube.__doc__)



#   Copyright 2011 Paul Voccio
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Launchpad Plugin"""

from launchpadlib.launchpad import Launchpad as LP

from pyhole import plugin
from pyhole import utils


class Launchpad(plugin.Plugin):
    """Provide access to the Launchpad API"""

    def __init__(self, irc):
        self.irc = irc
        self.name = self.__class__.__name__
        self.launchpad = LP.login_anonymously("pyhole", "production",
                utils.get_directory(self.name))

    @plugin.hook_add_command("lbugs")
    @utils.spawn
    def lbugs(self, params=None, **kwargs):
        """Launchpad bugs for a team (ex: .lbugs <project> <team>|<user>)"""
        if params:
            project, team = params.split(" ", 2)
            try:
                members = self.launchpad.people[team]
                proj = self.launchpad.projects[project]

                if len(members.members) < 2:
                    # Find a single member
                    self._find_bugs(members, proj)
                else:
                    # Find everyone on the team
                    for i, person in enumerate(members.members):
                        if i <= 4:
                            self._find_bugs(person, proj, False)
                        else:
                            self.irc.reply("[...] truncated last %d users" % (
                                    len(members.members) - i))
                            break
            except KeyError:
                self.irc.reply("Unable to find user '%s' in Launchpad" % team)
        else:
            self.irc.reply(self.lbugs.__doc__)

    @plugin.hook_add_keyword("lp")
    @utils.spawn
    def keyword_lp(self, params=None, **kwargs):
        """Retrieve Launchpad bug information (ex: LP12345)"""
        if params:
            params = utils.ensure_int(params)
            if not params:
                return

            try:
                bug = self.launchpad.bugs[params]
                task = bug.bug_tasks[len(bug.bug_tasks) - 1]

                self.irc.reply("LP %s [Status: %s, Assignee: %s] %s" % (
                        task.title, task.status,
                        self._find_name(task.assignee_link), bug.web_link))
            except Exception:
                return

    @plugin.hook_add_msg_regex("https?:\/\/bugs\.launchpad\.net\/.*\/\+bug")
    def _watch_for_lp_bug_url(self, params=None, **kwargs):
        """Watch for Launchpad bug URLs"""
        try:
            line = kwargs["full_message"].split("/")
            for i, word in enumerate(line):
                if word == "+bug":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_lp(bug_id)
        except TypeError:
            return

    @plugin.hook_add_msg_regex("https?:\/\/bugs\.launchpad\.net\/bugs")
    def _watch_for_short_lp_bug_url(self, params=None, **kwargs):
        """Watch for short Launchpad bug URLs"""
        try:
            line = kwargs["full_message"].split("/")
            for i, word in enumerate(line):
                if word == "bugs":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_lp(bug_id)
        except TypeError:
            return

    def _find_name(self, user):
        """Lookup a Launchpad user's display name"""
        try:
            return self.launchpad.people[user].display_name
        except ValueError:
            return "None"

    def _find_bugs(self, person, project, single=True):
        """Lookup Launchpad bugs"""
        bugs = project.searchTasks(assignee=person)
        for i, bug in enumerate(bugs):
            if i <= 4:
                self.irc.reply("LP %s [Assignee: %s] %s" % (bug.title,
                        person.display_name, bug.web_link))
            else:
                self.irc.reply("[...] truncated last %d bugs" % (
                        len(bugs) - i))
                break

        if single and i < 1:
            self.irc.reply("No bugs found for %s" % (person.display_name))



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Administration Plugin"""

from pyhole import irc
from pyhole import plugin
from pyhole import utils


class Admin(plugin.Plugin):
    """Provide administration functionality"""

    @plugin.hook_add_command("help")
    def help(self, params=None, **kwargs):
        """Learn how to use active commands (ex: .help <command>)"""

        if params:
            doc = self._find_doc_string(params)
            if doc:
                self.irc.reply(doc)
            else:
                self.irc.reply("No help available for %s" % params)
        else:
            self.irc.reply(self.help.__doc__)
            self.irc.reply("Active Commands: %s" % irc.active_commands())
            self.irc.reply("Active Keywords: %s" % irc.active_keywords())

    @plugin.hook_add_command("version")
    def version(self, params=None, **kwargs):
        """Display the current version"""
        self.irc.reply(self.irc.version)

    @plugin.hook_add_command("reload")
    @utils.admin
    def reload(self, params=None, **kwargs):
        """Reload all plugins"""
        self.irc.load_plugins(reload_plugins=True)
        self.irc.reply("Loaded Plugins: %s" % irc.active_plugins())

    @plugin.hook_add_command("op")
    @utils.admin
    def op(self, params=None, **kwargs):
        """Op a user (ex: .op <channel> <nick>)"""
        if params:
            self.irc.op_user(params)
        else:
            self.irc.reply(self.op.__doc__)

    @plugin.hook_add_command("deop")
    @utils.admin
    def deop(self, params=None, **kwargs):
        """De-op a user (ex: .deop <channel> <nick>)"""
        if params:
            self.irc.deop_user(params)
        else:
            self.irc.reply(self.deop.__doc__)

    @plugin.hook_add_command("nick")
    @utils.admin
    def nick(self, params=None, **kwargs):
        """Change IRC nick (ex: .nick <nick>)"""
        if params:
            self.irc.set_nick(params)
        else:
            self.irc.reply(self.nick.__doc__)

    @plugin.hook_add_command("join")
    @utils.admin
    def join(self, params=None, **kwargs):
        """Join a channel (ex: .join #channel [<key>])"""
        if params:
            self.irc.join_channel(params)
        else:
            self.irc.reply(self.join.__doc__)

    @plugin.hook_add_command("part")
    @utils.admin
    def part(self, params=None, **kwargs):
        """Part a channel (ex: .part <channel>)"""
        if params:
            self.irc.part_channel(params)
        else:
            self.irc.reply(self.part.__doc__)

    @plugin.hook_add_command("say")
    @utils.admin
    def say(self, params=None, **kwargs):
        """Send a PRIVMSG (ex: .say <channel>|<nick> message)"""
        if params:
            (target, msg) = params.split(" ", 1)
            self.irc.privmsg(target, msg)
        else:
            self.irc.reply(self.say.__doc__)

    def _find_doc_string(self, params):
        """Find the doc string for a plugin, command or keyword hook"""
        for p in plugin.active_plugin_classes():
            if p.__name__.upper() == params.upper():
                return p.__doc__

        for _, cmd_hook, cmd in plugin.hook_get_commands():
            if cmd.upper() == params.upper():
                return cmd_hook.__doc__

        for _, kw_hook, kw in plugin.hook_get_keywords():
            if kw.upper() == params.upper():
                return kw_hook.__doc__

        return None



#   Copyright 2011 Paul Voccio
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Taskwarrior Plugin"""

import subprocess

from pyhole import plugin


class Taskwarrior(plugin.Plugin):
    """Provide access to Taskwarrior"""

    @plugin.hook_add_command("task")
    def task(self, params=None, **kwargs):
        """List and add tasks (ex: .task <list|add|done>)"""
        if params:
            verb = params.split(" ", 1)[0]
            source = self.irc.source.split("!")[0]
            params = "%s project:%s" % (params, source)

            for command in ["list", "add", "done"]:
                if verb == command:
                    self._run(params)
        else:
            self.irc.reply(self.task.__doc__)

    def _run(self, params):
        """Execute the Taskwarrior binary"""
        try:
            p = subprocess.Popen(["task"] + params.split(" "),
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            result, error = p.communicate()

            self.irc.reply(result)
        except Exception:
            self.irc.reply("Unable to load Taskwarrior")



#   Copyright 2011 Chris Behrens
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Plugin Library"""

import functools
import os
import sys

import log
import utils


LOG = log.get_logger()


def _reset_variables():
    """Local function to init some variables that are common between
    load and reload
    """
    global _plugin_instances
    global _plugin_hooks

    _plugin_instances = []
    _plugin_hooks = {}

    for x in _hook_names:
        _plugin_hooks[x] = []


def hook_add(hookname, arg):
    """Generic decorator to add hooks.  Generally, this is not called
    directly by plugins.  Decorators that plugins use are automatically
    generated below with the setattrs you'll see
    """
    def wrap(f):
        setattr(f, "_is_%s_hook" % hookname, True)
        f._hook_arg = arg
        return f

    return wrap


def hook_get(hookname):
    """Function to return the list of hooks of a particular type.  Genearlly
    this is not called directly.  Callers tend to use the dynamically
    generated calls 'hook_get_*' that are created below with the setattrs
    """
    return _plugin_hooks[hookname]


def active_get(hookname):
    """Function to return the list of hook arguments.  Genearlly
    this is not called directly.  Callers tend to use the dynamically
    generated calls 'active_get_*' that are created below with the
    setattrs
    """
    return [x[2] for x in _plugin_hooks[hookname]]

_hook_names = ["keyword", "command", "msg_regex"]
_reset_variables()
_this_mod = sys.modules[__name__]

for x in _hook_names:
    # Dynamically create the decorators and functions for various hooks
    setattr(_this_mod, "hook_add_%s" % x, functools.partial(hook_add, x))
    setattr(_this_mod, "hook_get_%ss" % x, functools.partial(hook_get, x))
    setattr(_this_mod, "active_%ss" % x, functools.partial(active_get, x))


class PluginMetaClass(type):
    """The metaclass that makes all of the plugin magic work.  All subclassing
    gets caught here, which we can use to have plugins automagically
    register themselves
    """

    def __init__(cls, name, bases, attrs):
        """Catch subclassing.  If the class doesn't yet have _plugin_classes,
        it means it's the Plugin class itself, otherwise it's a class
        that's been subclassed from Plugin (ie, a real plugin class)
        """
        if not hasattr(cls, "_plugin_classes"):
            cls._plugin_classes = []
        else:
            cls._plugin_classes.append(cls)
        cls.__name__ = name


class Plugin(object):
    """The class that all plugin classes should inherit from"""

    __metaclass__ = PluginMetaClass

    def __init__(self, irc, *args, **kwargs):
        """Default constructor for Plugin.  Stores the IRC instance, etc"""
        self.irc = irc
        self.name = self.__class__.__name__


def _init_plugins(*args, **kwargs):
    """Create instances of the plugin classes and create a cache
    of their hook functions
    """
    for cls in Plugin._plugin_classes:
        # Create instance of 'p'
        instance = cls(*args, **kwargs)
        # Store the instance
        _plugin_instances.append(instance)

        # Setup _keyword_hooks by looking at all of the attributes
        # in the class and finding the ones that have a _is_*_hook
        # attribute
        for attr_name in dir(instance):
            attr = getattr(instance, attr_name)

            for hook_key in _hook_names:
                if getattr(attr, "_is_%s_hook" % hook_key, False):
                    hook_arg = getattr(attr, "_hook_arg", None)
                    # Append (module, method, arg) tuple
                    _plugin_hooks[hook_key].append((attr.__module__, attr,
                            hook_arg))


def load_user_plugin(plugin, *args, **kwargs):
    """Load a user plugin"""
    sys.path.append(utils.get_home_directory() + "plugins")
    user_plugins = os.listdir(utils.get_directory("plugins"))

    for user_plugin in user_plugins:
        if user_plugin.endswith(".py"):
            user_plugin = user_plugin[:-3]
            if plugin == user_plugin:
                try:
                    __import__(plugin, globals(), locals(), [plugin])
                except Exception, exc:
                    LOG.error(exc)


def load_plugins(*args, **kwargs):
    """Module function that loads plugins from a particular directory"""
    config = utils.get_config()
    plugin_names = config.get("plugins", type="list")

    for plugin_name in plugin_names:
        load_user_plugin(plugin_name, *args, **kwargs)

        try:
            __import__("pyhole.plugins", globals(), locals(), [plugin_name])
        except Exception, exc:
            LOG.error(exc)

    _init_plugins(*args, **kwargs)


def reload_plugins(*args, **kwargs):
    """Module function that'll reload all of the plugins"""
    config = utils.get_config()

    # When the modules are reloaded, the meta class will append
    # all of the classes again, so we need to make sure this is empty
    Plugin._plugin_classes = []
    _reset_variables()

    # Now reload all of the plugins
    plugins_to_reload = []
    plugindir = "pyhole.plugins"

    # Reload existing plugins
    for mod, val in sys.modules.items():
        if plugindir in mod and val and mod != plugindir:
            mod_file = val.__file__
            if not os.path.isfile(mod_file):
                continue
            for p in config.get("plugins", type="list"):
                if plugindir + "." + p == mod:
                    plugins_to_reload.append(mod)

    for plugin in plugins_to_reload:
        try:
            reload(sys.modules[plugin])
        except Exception, exc:
            LOG.error(exc)

    # Load new plugins
    load_plugins(*args, **kwargs)


def active_plugins():
    """Get the loaded plugin names"""
    return [x.__name__ for x in Plugin._plugin_classes]


def active_plugin_classes():
    """Get the loaded plugin classes"""
    return Plugin._plugin_classes



#   Copyright 2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import setuptools

from pyhole import version


setuptools.setup(
    name="irc-pyhole",
    version=version.version_hash(),
    author="Josh Kearney",
    author_email="josh@jk0.org",
    description="A modular IRC bot for Python developers.",
    license="Apache License, Version 2.0",
    url="http://docs.pyhole.org",
    packages=["pyhole", "pyhole.plugins"],
    install_requires=[
        "eventlet",
        "beautifulsoup==3.2.0",
        "launchpadlib",
        "pywapi"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": ["pyhole = pyhole.irc:main"]
    }
)



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Redmine Plugin"""

try:
    import json
except ImportError:
    import simplejson as json

from pyhole import plugin
from pyhole import utils


class Redmine(plugin.Plugin):
    """Provide access to the Redmine API"""

    def __init__(self, irc):
        self.irc = irc
        self.name = self.__class__.__name__
        self.disabled = False

        try:
            self.redmine = utils.get_config("Redmine")
            self.redmine_domain = self.redmine.get("domain")
            self.redmine_key = self.redmine.get("key")
            self.redmine_url = "https://%s:password@%s" % (
                    self.redmine_key, self.redmine_domain)
        except Exception:
            self.disabled = True

    @plugin.hook_add_command("rbugs")
    @utils.spawn
    def rbugs(self, params=None, **kwargs):
        """Redmine bugs for a user (ex: .rbugs <login>)"""
        if params and not self.disabled:
            login = params.split(" ", 1)[0]
            user_id = self._find_user(login)

            i = 0
            issues = self._find_issues(user_id)
            for i, issue in enumerate(issues):
                if i <= 4:
                    self._find_issue(issue["id"])
                else:
                    self.irc.reply("[...] truncated last %d bugs" % (
                            len(issues) - i))
                    break
            else:
                if i <= 0:
                    self.irc.reply("No Redmine bugs found for '%s'" % login)
        else:
            self.irc.reply(self.rbugs.__doc__)

    @plugin.hook_add_keyword("rm")
    @utils.spawn
    def keyword_rm(self, params=None, **kwargs):
        """Retrieve Redmine bug information (ex: RM12345)"""
        if params and not self.disabled:
            params = utils.ensure_int(params)
            if params:
                self._find_issue(params)

    @plugin.hook_add_msg_regex("https?:\/\/redmine\..*/issues")
    def _watch_for_rm_bug_url(self, params=None, **kwargs):
        """Watch for Redmine bug URLs"""
        try:
            line = kwargs["full_message"].split("/")
            for i, word in enumerate(line):
                if word == "issues":
                    bug_id = line[i + 1].split(" ", 1)[0]
                    self.keyword_rm(bug_id)
        except TypeError:
            return

    def _find_issues(self, user_id):
        """Find all issues for a Redmine user"""
        url = "%s/issues.json?assigned_to_id=%s" % (
                self.redmine_url, user_id)
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        return json.loads(response.read())["issues"]

    def _find_user(self, login):
        """Find a specific Redmine user"""
        for user in self._find_users():
            if login == user["login"]:
                return user["id"]

        for user in self._find_users(100):
            if login == user["login"]:
                return user["id"]

    def _find_users(self, offset=None):
        """Find all Redmine users"""
        if offset:
            url = "%s/users.json?limit=100&offset=%d" % (
                    self.redmine_url, offset)
        else:
            url = "%s/users.json?limit=100" % self.redmine_url
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        return json.loads(response.read())["users"]

    def _find_issue(self, issue_id):
        """Find and display a Redmine issue"""
        url = "%s/issues/%s.json" % (self.redmine_url, issue_id)
        response = self.irc.fetch_url(url, self.name)
        if not response:
            return

        try:
            issue = json.loads(response.read())["issue"]
        except Exception:
            return

        self.irc.reply("RM %s #%s: %s [Status: %s, Assignee: %s] "
                "https://%s/issues/show/%s" % (
                issue["tracker"]["name"], issue["id"], issue["subject"],
                issue["status"]["name"],
                issue.get("assigned_to", {}).get("name", "N/A"),
                self.redmine_domain, issue["id"]))



#   Copyright 2010-2011 Josh Kearney
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Pyhole Configuration Library"""

from __future__ import with_statement

import ConfigParser
import os
import sys

import utils


class Config(object):
    """A configuration object."""

    def __init__(self, config, section):
        self.config = os.path.abspath(config)
        self.config_parser = ConfigParser.ConfigParser()
        self.section = section

        try:
            with open(self.config) as conf_file:
                self.config_parser.readfp(conf_file)
        except IOError:
            print "Unable to load configuration file: %s" % self.config
            utils.generate_config()
            sys.exit(1)

    def __str__(self):
        """Make the config object readable for logging."""
        return self.section

    def sections(self):
        """Return a list of sections."""
        return self.config_parser.sections()

    def get(self, option, **kwargs):
        """Retrieve configuration values."""
        _type = kwargs.get("type", "str")

        try:
            if _type == "int":
                return self.config_parser.getint(self.section, option)
            elif _type == "bool":
                return self.config_parser.getboolean(self.section, option)
            elif _type == "list":
                return self.config_parser.get(self.section, option).split(", ")
            else:
                return self.config_parser.get(self.section, option)
        except ConfigParser.NoOptionError:
            if "default" in kwargs:
                return kwargs["default"]

            print "Unable to locate '%s' in %s" % (option, self.config)
            print "[%s]" % self.section
            print "%s: value" % option
            sys.exit(1)
